# ------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
# ------------------------------------------------------------------------------
import pytest

from tweedledum.bool_function_compiler.function_parser import ParseError
from tweedledum.bool_function_compiler.expression_parser import ExpressionParser


def _get_node(parser, symbol):
    _, signals_ = parser._symbol_table[symbol]
    return parser._logic_network.get_node(signals_[0])


def test_ordered():
    parser = ExpressionParser(
        "((A & C) | (B & D)) & ~(C & D)", var_order=["A", "B", "C", "D"]
    )
    assert _get_node(parser, "A") == parser._logic_network.pi_at(0)
    assert _get_node(parser, "B") == parser._logic_network.pi_at(1)
    assert _get_node(parser, "C") == parser._logic_network.pi_at(2)
    assert _get_node(parser, "D") == parser._logic_network.pi_at(3)


def test_fail_ordered():
    with pytest.raises(ParseError, match=r".*\{'C'\}*"):
        ExpressionParser("((A & C) | (B & D)) & ~(C & D)", var_order=["A", "B", "D"])


def test_unordered():
    parser = ExpressionParser("((A & C) | (B & D)) & ~(C & D)")
    assert _get_node(parser, "A") == parser._logic_network.pi_at(0)
    assert _get_node(parser, "B") == parser._logic_network.pi_at(2)
    assert _get_node(parser, "C") == parser._logic_network.pi_at(1)
    assert _get_node(parser, "D") == parser._logic_network.pi_at(3)



# ------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
# ------------------------------------------------------------------------------
import operator
import pytest
import random
from random import randrange

from tweedledum.bool_function_compiler import BitVec

random.seed(1337)


@pytest.fixture
def create_string_bitvectors():
    bvs = list()
    for i in range(1, 12):
        value = randrange((1 << i), (1 << (i + 1)))
        string = "{0:b}".format(value)
        bvs.append((BitVec((1 << i), value), string))
    return bvs


def test_one_bit_str():
    zero_str = BitVec(1, "0")
    assert zero_str._length == 1
    assert zero_str._value == 0


def test_one_bit_str():
    zero_str = BitVec(1, "0")
    assert zero_str._length == 1
    assert zero_str._value == 0

    one_str = BitVec(1, "1")
    assert one_str._length == 1
    assert one_str._value == 1


def test_one_bit_value():
    zero = BitVec(1, 0)
    assert zero._length == 1
    assert zero._value == 0

    one = BitVec(1, 1)
    assert one._length == 1
    assert one._value == 1


def test_fail_value_type():
    with pytest.raises(TypeError, match=r".*interger or a bit string.*"):
        BitVec(1, 1.0)


def test_fail_size_str():
    with pytest.raises(TypeError, match=r".*length 2.*length 1.*"):
        BitVec(1, "00")


def test_fail_size_value():
    with pytest.raises(TypeError, match=r".*length 4.*length 1.*"):
        BitVec(1, 10)


def test_n_bit_strings(create_string_bitvectors):
    for _, string in create_string_bitvectors:
        bv = BitVec(len(string), string)
        assert bv._length == len(string)
        assert bv._value == int(string, 2)

        bv = BitVec(string)
        assert bv._length == len(string)
        assert bv._value == int(string, 2)


def test_getitem(create_string_bitvectors):
    for bv, _ in create_string_bitvectors:
        string = str(bv)
        last = len(string) - 1
        for i in range(len(string)):
            value = bv[i]._value
            assert bv[i]._length == 1
            # Remember that the string is little-endian!
            assert value == int(string[last - i])


def test_setitem(create_string_bitvectors):
    for bv, _ in create_string_bitvectors:
        string = str(bv)
        result = BitVec(bv._length, 0)
        last = len(string) - 1
        for i in range(len(string)):
            result[i] = int(string[last - i])
        assert bv._length == result._length
        assert bv._value == result._value


def test_getslice(create_string_bitvectors):
    # Limit the length
    for bv, string in create_string_bitvectors[0:32]:
        for i in range(1, len(string) + 1):
            for j in range(len(string)):
                if i <= j:
                    with pytest.raises(ValueError, match=r".*i > j.*"):
                        bv[i:j]
                    continue
                result = bv[i:j]
                expected = string[len(string) - i : len(string) - j]
                assert result._length == len(expected)
                assert result._value == int(expected, 2)


def test_setslice(create_string_bitvectors):
    for bv, _ in create_string_bitvectors[3:]:
        string = str(bv)
        result = BitVec(bv._length, (1 << bv._length - 1) - 1)
        n = int(bv._length / 4)
        for i in range(0, len(string), n):
            part = string[i : i + n]
            result[len(string) - i : len(string) - (i + n)] = BitVec(len(part), part)
        assert result._length == bv._length
        assert result._value == bv._value


@pytest.fixture
def create_bv_operand_pairs():
    operand_pairs = list()
    for i in range(1, 12):
        left = randrange((1 << i), (1 << (i + 1)))
        right = randrange((1 << i), (1 << (i + 1)))
        operand_pairs.append((BitVec((1 << i), left), BitVec((1 << i), right)))
    return operand_pairs


def test_binops(create_bv_operand_pairs):
    for op in [operator.and_, operator.or_, operator.xor]:
        for left, right in create_bv_operand_pairs:
            result = op(left, right)
            iresult = op(right, left)
            assert result._length == left._length
            assert result._value == op(left._value, right._value)
            assert iresult._value == op(left._value, right._value)



# ------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
# ------------------------------------------------------------------------------
from tweedledum.bool_function_compiler import BitVec, BoolFunction

def constant() -> BitVec(1):
    return BitVec(1, '1')

def constant_2bit() -> BitVec(2):
    return BitVec(2, '10')

def constant_3bit() -> BitVec(3):
    return BitVec('101')

def constant_4bit() -> BitVec(4):
    return BitVec(4)

def identity(a : BitVec(1)) -> BitVec(1):
    return a

def identity_2bit(a : BitVec(2)) -> BitVec(2):
    return a

def identity_not(a : BitVec(2)) -> (BitVec(2), BitVec(2)):
    return a, ~a

def bool_not(a : BitVec(1)) -> BitVec(1):
    return not a

def bit_not(a : BitVec(1)) -> BitVec(1):
    return ~a

def bit_not_2bit(a : BitVec(2)) -> BitVec(2):
    return ~a

def bool_and(a, b : BitVec(1)) -> BitVec(1):
    return a and b

def bit_and(a, b : BitVec(1)) -> BitVec(1):
    return a & b

def bit_and_2bit(a, b : BitVec(2)) -> BitVec(2):
    return a & b

def bool_or(a, b : BitVec(1)) -> BitVec(1):
    return a or b

def bit_or(a, b : BitVec(1)) -> BitVec(1):
    return a | b

def bit_or_2bit(a, b : BitVec(2)) -> BitVec(2):
    return a | b

def bit_xor(a, b : BitVec(1)) -> BitVec(1):
    return a ^ b

def bit_xor_2bit(a, b : BitVec(2)) -> BitVec(2):
    return a ^ b

def eq(a, b : BitVec(1)) -> BitVec(1):
    return a == b

def eq_2bit(a, b : BitVec(2)) -> BitVec(1):
    return a == b

def ne(a, b : BitVec(1)) -> BitVec(1):
    return a != b

def ne_2bit(a, b : BitVec(2)) -> BitVec(1):
    return a != b

class TestBoolFunction:
    def test_constant_3bit(self):
        function = BoolFunction(constant_3bit)
        assert function._parameters_signature == []
        result = constant_3bit()
        assert result == BitVec(3, "101")

    def test_identity(self):
        function = BoolFunction(identity)
        assert function._parameters_signature == [(type(BitVec(1)), 1)]
        result = identity(BitVec(1, "0"))
        assert result == BitVec(1, "0")
        result = identity(BitVec(1, "1"))
        assert result == BitVec(1, "1")

    def test_identity_2bit(self):
        function = BoolFunction(identity_2bit)
        assert function._parameters_signature == [(type(BitVec(2)), 2)]
        for a in range(4):
            tmp = BitVec(2, a)
            result = identity(tmp)
            assert result == tmp

    def test_bool_not(self):
        function = BoolFunction(bool_not)
        assert function._parameters_signature == [(type(BitVec(1)), 1)]
        for a in range(2):
            tmp = BitVec(1, a)
            result = bool_not(tmp)
            assert result is not bool(tmp)

    def test_bit_not(self):
        function = BoolFunction(bit_not)
        assert function._parameters_signature == [(type(BitVec(1)), 1)]
        for a in range(2):
            tmp = BitVec(1, a)
            result = bit_not(tmp)
            assert result == ~tmp

    def test_bit_not_2bit(self):
        function = BoolFunction(bit_not_2bit)
        assert function._parameters_signature == [(type(BitVec(2)), 2)]
        for a in range(4):
            tmp = BitVec(2, a)
            result = bit_not_2bit(tmp)
            assert result == ~tmp

    def test_bool_and(self):
        function = BoolFunction(bool_and)
        assert function._parameters_signature == [
            (type(BitVec(1)), 1),
            (type(BitVec(1)), 1),
        ]
        for a in range(2):
            for b in range(2):
                result = bool_and(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) and BitVec(1, b)
                assert result == tmp

    def test_bit_and(self):
        function = BoolFunction(bit_and)
        assert function._parameters_signature == [
            (type(BitVec(1)), 1),
            (type(BitVec(1)), 1),
        ]
        for a in range(2):
            for b in range(2):
                result = bit_and(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) & BitVec(1, b)
                assert result == tmp

    def test_bit_and_2bit(self):
        function = BoolFunction(bit_and_2bit)
        assert function._parameters_signature == [
            (type(BitVec(2)), 2),
            (type(BitVec(2)), 2),
        ]
        for a in range(4):
            for b in range(4):
                result = bit_and_2bit(BitVec(2, a), BitVec(2, b))
                tmp = BitVec(2, a) & BitVec(2, b)
                assert result == tmp

    def test_bool_or(self):
        function = BoolFunction(bool_or)
        assert function._parameters_signature == [
            (type(BitVec(1)), 1),
            (type(BitVec(1)), 1),
        ]
        for a in range(2):
            for b in range(2):
                result = bool_or(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) or BitVec(1, b)
                assert result == tmp

    def test_bit_or(self):
        function = BoolFunction(bit_or)
        assert function._parameters_signature == [
            (type(BitVec(1)), 1),
            (type(BitVec(1)), 1),
        ]
        for a in range(2):
            for b in range(2):
                result = bit_or(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) | BitVec(1, b)
                assert result == tmp

    def test_bit_or_2bit(self):
        function = BoolFunction(bit_or_2bit)
        assert function._parameters_signature == [
            (type(BitVec(2)), 2),
            (type(BitVec(2)), 2),
        ]
        for a in range(4):
            for b in range(4):
                result = bit_or_2bit(BitVec(2, a), BitVec(2, b))
                tmp = BitVec(2, a) | BitVec(2, b)
                assert result == tmp

    def test_bit_xor(self):
        function = BoolFunction(bit_xor)
        assert function._parameters_signature == [
            (type(BitVec(1)), 1),
            (type(BitVec(1)), 1),
        ]
        for a in range(2):
            for b in range(2):
                result = bit_xor(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) ^ BitVec(1, b)
                assert result == tmp

    def test_bit_xor_2bit(self):
        function = BoolFunction(bit_xor_2bit)
        assert function._parameters_signature == [
            (type(BitVec(2)), 2),
            (type(BitVec(2)), 2),
        ]
        for a in range(4):
            for b in range(4):
                result = bit_xor_2bit(BitVec(2, a), BitVec(2, b))
                tmp = BitVec(2, a) ^ BitVec(2, b)
                assert result == tmp

    def test_eq(self):
        function = BoolFunction(eq)
        assert function._parameters_signature == [
            (type(BitVec(1)), 1),
            (type(BitVec(1)), 1),
        ]
        for a in range(2):
            for b in range(2):
                result = eq(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) == BitVec(1, b)
                assert result == tmp

    def test_bit_eq_2bit(self):
        function = BoolFunction(eq_2bit)
        assert function._parameters_signature == [
            (type(BitVec(2)), 2),
            (type(BitVec(2)), 2),
        ]
        for a in range(4):
            for b in range(4):
                result = eq_2bit(BitVec(2, a), BitVec(2, b))
                tmp = BitVec(2, a) == BitVec(2, b)
                assert result == tmp

    def test_ne(self):
        function = BoolFunction(ne)
        assert function._parameters_signature == [
            (type(BitVec(1)), 1),
            (type(BitVec(1)), 1),
        ]
        for a in range(2):
            for b in range(2):
                result = ne(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) != BitVec(1, b)
                assert result == tmp

    def test_bit_ne_2bit(self):
        function = BoolFunction(ne_2bit)
        assert function._parameters_signature == [
            (type(BitVec(2)), 2),
            (type(BitVec(2)), 2),
        ]
        for a in range(4):
            for b in range(4):
                result = ne_2bit(BitVec(2, a), BitVec(2, b))
                tmp = BitVec(2, a) != BitVec(2, b)
                assert result == tmp


class TestBoolFunctionSimulation:
    def test_constant(self):
        function = BoolFunction(constant)
        result = function.simulate()
        assert result == BitVec(1, "1")

    def test_constant_2bit(self):
        function = BoolFunction(constant_2bit)
        result = function.simulate()
        assert result == BitVec("10")

    def test_constant_3bit(self):
        function = BoolFunction(constant_3bit)
        result = function.simulate()
        assert result == BitVec(3, 5)

    def test_constant_4bit(self):
        function = BoolFunction(constant_4bit)
        result = function.simulate()
        assert result == BitVec("0000")

    def test_identity(self):
        function = BoolFunction(identity)
        result = function.simulate(BitVec(1, "0"))
        assert result == BitVec(1, "0")
        result = function.simulate(BitVec(1, "1"))
        assert result == BitVec(1, "1")

    def test_identity_2bit(self):
        function = BoolFunction(identity_2bit)
        for a in range(4):
            tmp = BitVec(2, a)
            result = function.simulate(tmp)
            assert result == tmp

    def test_identity_not(self):
        function = BoolFunction(identity_not)
        for a in range(4):
            tmp = BitVec(2, a)
            result = function.simulate(tmp)
            assert result == (tmp, ~tmp)

    def test_bool_not(self):
        function = BoolFunction(bool_not)
        for a in range(2):
            tmp = BitVec(1, a)
            result = function.simulate(tmp)
            assert result == ~tmp

    def test_bit_not(self):
        function = BoolFunction(bit_not)
        for a in range(2):
            tmp = BitVec(1, a)
            result = function.simulate(tmp)
            assert result == ~tmp

    def test_bit_not_2bit(self):
        function = BoolFunction(bit_not_2bit)
        for a in range(4):
            tmp = BitVec(2, a)
            result = function.simulate(tmp)
            assert result == ~tmp

    def test_bool_and(self):
        function = BoolFunction(bool_and)
        for a in range(2):
            for b in range(2):
                result = function.simulate(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) and BitVec(1, b)
                assert result == tmp

    def test_bit_and(self):
        function = BoolFunction(bit_and)
        for a in range(2):
            for b in range(2):
                result = function.simulate(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) & BitVec(1, b)
                assert result == tmp

    def test_bit_and_2bit(self):
        function = BoolFunction(bit_and_2bit)
        for a in range(4):
            for b in range(4):
                result = function.simulate(BitVec(2, a), BitVec(2, b))
                tmp = BitVec(2, a) & BitVec(2, b)
                assert result == tmp

    def test_bool_or(self):
        function = BoolFunction(bool_or)
        for a in range(2):
            for b in range(2):
                result = function.simulate(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) or BitVec(1, b)
                assert result == tmp

    def test_bit_or(self):
        function = BoolFunction(bit_or)
        for a in range(2):
            for b in range(2):
                result = function.simulate(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) | BitVec(1, b)
                assert result == tmp

    def test_bit_or_2bit(self):
        function = BoolFunction(bit_or_2bit)
        for a in range(4):
            for b in range(4):
                result = function.simulate(BitVec(2, a), BitVec(2, b))
                tmp = BitVec(2, a) | BitVec(2, b)
                assert result == tmp

    def test_bit_xor(self):
        function = BoolFunction(bit_xor)
        for a in range(2):
            for b in range(2):
                result = function.simulate(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) ^ BitVec(1, b)
                assert result == tmp

    def test_bit_xor_2bit(self):
        function = BoolFunction(bit_xor_2bit)
        for a in range(4):
            for b in range(4):
                result = function.simulate(BitVec(2, a), BitVec(2, b))
                tmp = BitVec(2, a) ^ BitVec(2, b)
                assert result == tmp

    def test_eq(self):
        function = BoolFunction(eq)
        assert function._parameters_signature == [
            (type(BitVec(1)), 1),
            (type(BitVec(1)), 1),
        ]
        for a in range(2):
            for b in range(2):
                result = function.simulate(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) == BitVec(1, b)
                assert result == tmp

    def test_bit_eq_2bit(self):
        function = BoolFunction(eq_2bit)
        assert function._parameters_signature == [
            (type(BitVec(2)), 2),
            (type(BitVec(2)), 2),
        ]
        for a in range(4):
            for b in range(4):
                result = function.simulate(BitVec(2, a), BitVec(2, b))
                tmp = BitVec(2, a) == BitVec(2, b)
                assert result == tmp

    def test_ne(self):
        function = BoolFunction(ne)
        assert function._parameters_signature == [
            (type(BitVec(1)), 1),
            (type(BitVec(1)), 1),
        ]
        for a in range(2):
            for b in range(2):
                result = function.simulate(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) != BitVec(1, b)
                assert result == tmp

    def test_bit_ne_2bit(self):
        function = BoolFunction(ne_2bit)
        assert function._parameters_signature == [
            (type(BitVec(2)), 2),
            (type(BitVec(2)), 2),
        ]
        for a in range(4):
            for b in range(4):
                result = function.simulate(BitVec(2, a), BitVec(2, b))
                tmp = BitVec(2, a) != BitVec(2, b)
                assert result == tmp


class TestBoolFunctionFullSimulation:
    def test_identity(self):
        function = BoolFunction(identity)
        function.simulate_all()
        assert function.num_inputs() == 1
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 1
        assert function.num_output_bits() == 1
        assert str(function.truth_table(output_bit=0)) == "10"

        result = function.simulate(BitVec(1, "0"))
        assert result == BitVec(1, "0")
        result = function.simulate(BitVec(1, "1"))
        assert result == BitVec(1, "1")

    def test_identity_2bit(self):
        function = BoolFunction(identity_2bit)
        function.simulate_all()
        assert function.num_inputs() == 1
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 2
        assert function.num_output_bits() == 2
        assert str(function.truth_table(output_bit=0)) == "1010"
        assert str(function.truth_table(output_bit=1)) == "1100"
        for a in range(4):
            tmp = BitVec(2, a)
            result = function.simulate(tmp)
            assert result == tmp

    def test_not(self):
        function = BoolFunction(bool_not)
        function.simulate_all()
        assert function.num_inputs() == 1
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 1
        assert function.num_output_bits() == 1
        assert str(function.truth_table(output_bit=0)) == "01"
        for a in range(2):
            tmp = BitVec(1, a)
            result = function.simulate(tmp)
            assert result == ~tmp

    def test_not_2bit(self):
        function = BoolFunction(bit_not_2bit)
        function.simulate_all()
        assert function.num_inputs() == 1
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 2
        assert function.num_output_bits() == 2
        assert str(function.truth_table(output_bit=0)) == "0101"
        assert str(function.truth_table(output_bit=1)) == "0011"
        for a in range(4):
            tmp = BitVec(2, a)
            result = function.simulate(tmp)
            assert result == ~tmp

    def test_and(self):
        function = BoolFunction(bool_and)
        function.simulate_all()
        assert function.num_inputs() == 2
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 2
        assert function.num_output_bits() == 1
        assert str(function.truth_table(output_bit=0)) == "1000"
        for a in range(2):
            for b in range(2):
                result = function.simulate(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) and BitVec(1, b)
                assert result == tmp

    def test_and_2bit(self):
        function = BoolFunction(bit_and_2bit)
        function.simulate_all()
        assert function.num_inputs() == 2
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 4
        assert function.num_output_bits() == 2
        output0 = BitVec(16, 0xAAAA) & BitVec(16, 0xF0F0)
        output1 = BitVec(16, 0xCCCC) & BitVec(16, 0xFF00)
        assert str(function.truth_table(output_bit=0)) == str(output0)
        assert str(function.truth_table(output_bit=1)) == str(output1)
        for a in range(4):
            for b in range(4):
                result = function.simulate(BitVec(2, a), BitVec(2, b))
                tmp = BitVec(2, a) & BitVec(2, b)
                assert result == tmp

    def test_or(self):
        function = BoolFunction(bool_or)
        function.simulate_all()
        assert function.num_inputs() == 2
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 2
        assert function.num_output_bits() == 1
        assert str(function.truth_table(output_bit=0)) == "1110"
        for a in range(2):
            for b in range(2):
                result = function.simulate(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) or BitVec(1, b)
                assert result == tmp

    def test_or_2bit(self):
        function = BoolFunction(bit_or_2bit)
        function.simulate_all()
        assert function.num_inputs() == 2
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 4
        assert function.num_output_bits() == 2
        output0 = BitVec(16, 0xAAAA) | BitVec(16, 0xF0F0)
        output1 = BitVec(16, 0xCCCC) | BitVec(16, 0xFF00)
        assert str(function.truth_table(output_bit=0)) == str(output0)
        assert str(function.truth_table(output_bit=1)) == str(output1)
        for a in range(4):
            for b in range(4):
                result = function.simulate(BitVec(2, a), BitVec(2, b))
                tmp = BitVec(2, a) | BitVec(2, b)
                assert result == tmp

    def test_xor_2bit(self):
        function = BoolFunction(bit_xor_2bit)
        function.simulate_all()
        assert function.num_inputs() == 2
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 4
        assert function.num_output_bits() == 2
        output0 = BitVec(16, 0xAAAA) ^ BitVec(16, 0xF0F0)
        output1 = BitVec(16, 0xCCCC) ^ BitVec(16, 0xFF00)
        assert str(function.truth_table(output_bit=0)) == str(output0)
        assert str(function.truth_table(output_bit=1)) == str(output1)
        for a in range(4):
            for b in range(4):
                result = function.simulate(BitVec(2, a), BitVec(2, b))
                tmp = BitVec(2, a) ^ BitVec(2, b)
                assert result == tmp


class TestBoolFunctionExpressionConstructor:
    def test_identity(self):
        function = BoolFunction.from_expression("x")
        function.simulate_all()
        assert function.num_inputs() == 1
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 1
        assert function.num_output_bits() == 1
        assert str(function.truth_table(output_bit=0)) == "10"

    def test_not(self):
        function = BoolFunction.from_expression("~x")
        function.simulate_all()
        assert function.num_inputs() == 1
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 1
        assert function.num_output_bits() == 1
        assert str(function.truth_table(output_bit=0)) == "01"

    def test_and(self):
        function = BoolFunction.from_expression("x & b")
        function.simulate_all()
        assert function.num_inputs() == 2
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 2
        assert function.num_output_bits() == 1
        assert str(function.truth_table(output_bit=0)) == "1000"

    def test_or(self):
        function = BoolFunction.from_expression("x | b")
        function.simulate_all()
        assert function.num_inputs() == 2
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 2
        assert function.num_output_bits() == 1
        assert str(function.truth_table(output_bit=0)) == "1110"

    def test_xor(self):
        function = BoolFunction.from_expression("x ^ b")
        function.simulate_all()
        assert function.num_inputs() == 2
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 2
        assert function.num_output_bits() == 1
        assert str(function.truth_table(output_bit=0)) == "0110"

    def test_de_morgan(self):
        function = BoolFunction.from_expression("~(~(x0 | x1) ^ (~x0 & ~x1))")
        function.simulate_all()
        assert function.num_inputs() == 2
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 2
        assert function.num_output_bits() == 1
        assert str(function.truth_table(output_bit=0)) == "1111"


class TestBoolFunctionTTConstructor:
    def test_identity(self):
        function = BoolFunction.from_truth_table("10")
        assert function.num_inputs() == 1
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 1
        assert function.num_output_bits() == 1
        result = function.simulate(BitVec(1, "0"))
        assert result == BitVec(1, "0")
        result = function.simulate(BitVec(1, "1"))
        assert result == BitVec(1, "1")

    def test_identity_2bit(self):
        function = BoolFunction.from_truth_table(["1010", "1100"])
        assert function.num_inputs() == 2
        assert function.num_outputs() == 2
        assert function.num_input_bits() == 2
        assert function.num_output_bits() == 2
        for a in range(4):
            tmp = BitVec(2, a)
            result = function.simulate(tmp[0], tmp[1])
            assert result == (tmp[0], tmp[1])

    def test_not(self):
        function = BoolFunction.from_truth_table("01")
        assert function.num_inputs() == 1
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 1
        assert function.num_output_bits() == 1
        for a in range(2):
            tmp = BitVec(1, a)
            result = function.simulate(tmp)
            assert result == ~tmp

    def test_not_2bit(self):
        function = BoolFunction.from_truth_table(["0101", "0011"])
        assert function.num_inputs() == 2
        assert function.num_outputs() == 2
        assert function.num_input_bits() == 2
        assert function.num_output_bits() == 2
        for a in range(4):
            tmp = BitVec(2, a)
            result = function.simulate(tmp[0], tmp[1])
            assert result == (~tmp[0], ~tmp[1])

    def test_and(self):
        function = BoolFunction.from_truth_table("1000")
        assert function.num_inputs() == 2
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 2
        assert function.num_output_bits() == 1
        for a in range(2):
            for b in range(2):
                result = function.simulate(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) and BitVec(1, b)
                assert result == tmp

    def test_and_2bit(self):
        function = BoolFunction.from_truth_table(
            ["1010000010100000", "1100110000000000"]
        )
        assert function.num_inputs() == 4
        assert function.num_outputs() == 2
        assert function.num_input_bits() == 4
        assert function.num_output_bits() == 2
        for a in range(4):
            for b in range(4):
                a_bv = BitVec(2, a)
                b_bv = BitVec(2, b)
                result = function.simulate(a_bv[0], a_bv[1], b_bv[0], b_bv[1])
                tmp = a_bv & b_bv
                assert result == (tmp[0], tmp[1])

    def test_or(self):
        function = BoolFunction(bool_and)
        function = BoolFunction.from_truth_table("1110")
        assert function.num_inputs() == 2
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 2
        assert function.num_output_bits() == 1
        for a in range(2):
            for b in range(2):
                result = function.simulate(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) or BitVec(1, b)
                assert result == tmp

    def test_or_2bit(self):
        function = BoolFunction.from_truth_table(
            ["1111101011111010", "1111111111001100"]
        )
        assert function.num_inputs() == 4
        assert function.num_outputs() == 2
        assert function.num_input_bits() == 4
        assert function.num_output_bits() == 2
        for a in range(4):
            for b in range(4):
                a_bv = BitVec(2, a)
                b_bv = BitVec(2, b)
                result = function.simulate(a_bv[0], a_bv[1], b_bv[0], b_bv[1])
                tmp = a_bv | b_bv
                assert result == (tmp[0], tmp[1])

    def test_xor(self):
        function = BoolFunction(bool_and)
        function = BoolFunction.from_truth_table("0110")
        assert function.num_inputs() == 2
        assert function.num_outputs() == 1
        assert function.num_input_bits() == 2
        assert function.num_output_bits() == 1
        for a in range(2):
            for b in range(2):
                result = function.simulate(BitVec(1, a), BitVec(1, b))
                tmp = BitVec(1, a) ^ BitVec(1, b)
                assert result == tmp

    def test_xor_2bit(self):
        function = BoolFunction.from_truth_table(
            ["0101101001011010", "0011001111001100"]
        )
        assert function.num_inputs() == 4
        assert function.num_outputs() == 2
        assert function.num_input_bits() == 4
        assert function.num_output_bits() == 2
        for a in range(4):
            for b in range(4):
                a_bv = BitVec(2, a)
                b_bv = BitVec(2, b)
                result = function.simulate(a_bv[0], a_bv[1], b_bv[0], b_bv[1])
                tmp = a_bv ^ b_bv
                assert result == (tmp[0], tmp[1])



# -------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
# -------------------------------------------------------------------------------
from typing import Union

from qiskit.transpiler.basepasses import TransformationPass
from qiskit.dagcircuit import DAGCircuit

from tweedledum.qiskit import from_qiskit, to_qiskit
from tweedledum.ir import Circuit
from tweedledum.passes import phase_folding


class PhaseFolding(TransformationPass):
    """TODO"""

    def __init__(self):
        """PhaseFolding initializer."""
        super().__init__()

    def run(self, dag: Union[DAGCircuit, Circuit]) -> Union[DAGCircuit, Circuit]:
        """Run the PhaseFolding pass on a circuit.

        Args:
            dag (Union[DAGCircuit, Circuit]): A circuit to map.

        Returns:
            Union[DAGCircuit, Circuit]: An optimized circuit.
        """
        if isinstance(dag, DAGCircuit):
            circuit = from_qiskit(dag)
        else:
            circuit = dag

        opt_circuit = phase_folding(circuit)

        if isinstance(dag, DAGCircuit):
            return to_qiskit(opt_circuit)
        return opt_circuit



#-------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
#-------------------------------------------------------------------------------
from .gate_cancellation import GateCancellation
from .linear_resynth import LinearResynth
from .phase_folding import PhaseFolding
from .steiner_resynth import SteinerResynth



# -------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
# -------------------------------------------------------------------------------
import ast

from .bitvec import BitVec
from .function_parser import FunctionParser, ParseError


class ExpressionParser(FunctionParser):
    def __init__(self, source, var_order: list = None):
        self.var_order = var_order
        super().__init__(source)

    def visit_Module(self, node):
        """The full snippet should contain a single function"""
        if len(node.body) != 1 or not isinstance(node.body[0], ast.Expr):
            raise ParseError("Not an expression to parse")
        if self.var_order:
            self._create_ordered_vars(node)

        self._return_signature = [(self.types["BitVec"], 1)]
        expr_type, expr_signal = self.visit(node.body[0])
        if expr_type != self._return_signature[0]:
            raise ParseError("Expressions can only evaluate to a BitVec(1)")
        self._logic_network.create_po(expr_signal[0])

    def visit_Expr(self, node):
        return super().visit(node.value)

    def visit_Name(self, node):
        return self._create_name(node.id)

    def _create_name(self, name: str):
        if name not in self._symbol_table:
            self._parameters_signature.append((self.types["BitVec"], 1))
            self._symbol_table[name] = (
                (self.types["BitVec"], 1),
                [self._logic_network.create_pi()],
            )
        return self._symbol_table[name]

    def _create_ordered_vars(self, module):
        vars = set()
        for node in ast.walk(module):
            if isinstance(node, ast.Name):
                vars.add(node.id)
        if vars != set(self.var_order):
            raise ParseError(
                f"Missing variables in order list {vars - set(self.var_order)}"
            )
        for var in self.var_order:
            self._create_name(var)



#-------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
#-------------------------------------------------------------------------------
"""These examples should be handle by the classicalfunction compiler"""

from tweedledum.bool_function_compiler import BitVec 

def constant() -> BitVec(1):
    return '1'

def constant_2bit() -> BitVec(2):
    return '10'

def identity(a : BitVec(1)) -> BitVec(2):
    return a

def identity_2bit(a : BitVec(1)) -> BitVec(2):
    return a

#-------------------------------------------------------------------------------

# Reason: Boolean NOT operator cannot be applied to a variable with multiple 
# bits
def bool_not_2bit(a : BitVec(2)) -> BitVec(2):
    return not a

#-------------------------------------------------------------------------------


# -------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
# -------------------------------------------------------------------------------
from .bitvec import BitVec
from .bool_function import BoolFunction

from tweedledum.ir import Circuit
from tweedledum.operators import X, H
from tweedledum.synthesis import (
    lhrs_synth,
    pprm_synth,
    pkrm_synth,
    spectrum_synth,
    xag_synth,
)


_METHOD_TO_CALLABLE = {
    "lhrs": lhrs_synth,
    "pprm": pprm_synth,
    "pkrm": pkrm_synth,
    "spectrum": spectrum_synth,
    "xag": xag_synth,
}


def bitflip_circuit(f: BoolFunction, method: str, config: dict() = {}):
    r"""Synthesizes a Boolean Function as a bitflip oracle.

    A bitflip oracle is a quantum operator `Bf` specified by a Boolean function
    `f` for which the effect on all computational basis states is given by

       Bf : |x>|y>|0>^a --> |x>|y + f(x)>|0>^a

    where `+` is the logical exclusive-or operator and `a >= 0` corresponds to
    the number of extra qubits used to store intermediate results for the
    computation of `f(x)`, the so-called ancillae qubits.

    (For clarity, this example only showed a single output function)

    Args:
        f(BoolFunction): Boolean function to be synthesized
        method: Synthesis method ('lhrs', 'pprm', 'pprm', 'spectrum', or 'xag')
        config: Dictionary with configuration parameters for tweedledum
    """
    synthesizer = _METHOD_TO_CALLABLE.get(method)
    if synthesizer is None:
        raise ValueError(f"Unrecognized synthesis method: {method}")
    if method in ["pprm", "pkrm", "spectrum"]:
        if f.num_outputs() > 1:
            raise ValueError("TT based methods only work for single output functions")
        if f.num_inputs() > 16:
            raise ValueError(
                "TT based methods only work for functions with at most 16 inputs"
            )
        return synthesizer(f.truth_table(output_bit=0), config)
    return synthesizer(f.logic_network(), config)


def phaseflip_circuit(f: BoolFunction, method: str, config: dict() = {}):
    r"""Synthesizes a Boolean Function as a phaseflip oracle.

    A phaseflip oracle is a quantum operator `Pf` specified by a Boolean
    function `f` for which the effect on all computational basis states is given
    by

        Bf : |x>|0>^a --> (-1)^{f(x)}|x>|0>^a

    Note that you can easily construct a phase oracle from a bit oracle by
    sandwiching the controlled X gate on the result qubit by a X and H gate.
    """
    synthesizer = _METHOD_TO_CALLABLE.get(method)
    if synthesizer is None:
        raise ValueError(f"Unrecognized synthesis method: {method}")
    if method in ["pprm", "pkrm", "spectrum"]:
        if f.num_outputs() > 1:
            raise ValueError("TT based methods only work for single output functions")
        if f.num_inputs() > 16:
            raise ValueError(
                "TT based methods only work for functions with at most 16 inputs"
            )
        if f"{method}_synth" not in config:
            config[f"{method}_synth"] = {"phase_esop": True}
        else:
            config[f"{method}_synth"]["phase_esop"] = True
        return synthesizer(f.truth_table(output_bit=0), config)

    circuit = Circuit()
    num_qubits = f.num_inputs() + f.num_outputs()
    qubits = [circuit.create_qubit() for _ in range(num_qubits)]
    cbits = list()
    for i in range(f.num_inputs(), num_qubits):
        circuit.apply_operator(X(), [qubits[i]])
        circuit.apply_operator(H(), [qubits[i]])
    synthesizer(circuit, qubits, cbits, f.logic_network(), config)
    for i in range(f.num_inputs(), num_qubits):
        circuit.apply_operator(H(), [qubits[i]])
        circuit.apply_operator(X(), [qubits[i]])
    return circuit



# -------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
# -------------------------------------------------------------------------------
from typing import Optional, Union

from qiskit.dagcircuit import DAGCircuit
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.coupling import CouplingMap
from qiskit.transpiler.exceptions import TranspilerError

from tweedledum.ir import Circuit
from tweedledum.passes import bridge_map, jit_map, sabre_map
from tweedledum.qiskit import from_qiskit, to_qiskit
from tweedledum.target import Device


class Mapping(TransformationPass):
    """TODO"""

    def __init__(
        self,
        cmap_or_device: Union[list, CouplingMap, Device],
        method: str = "sabre",
        seed: Optional[int] = None,
        fake_run: bool = False,
    ):
        """Mapping initializer.

        Args:
            cmap_or_device (Union[list, CouplingMap, Device]): Directed graph
                represented as a list of edges, or a qiskit coupling map, or a
                tweedledum Device.
        """
        super().__init__()
        if isinstance(cmap_or_device, list):
            self.device = Device.from_edge_list(cmap_or_device)
        elif isinstance(cmap_or_device, CouplingMap):
            self.device = Device.from_edge_list(cmap_or_device.get_edges())
        elif isinstance(cmap_or_device, Device):
            self.device = cmap_or_device
        else:
            raise TranspilerError("Must pass a qiskit CouplingMap or tweedledum Device")
        self.method = method

    def run(self, dag: Union[DAGCircuit, Circuit]) -> Union[DAGCircuit, Circuit]:
        """Run the Mapping pass on `dag`.

        Args:
            dag (DAGCircuit): DAG to map.

        Returns:
            DAGCircuit: A mapped DAG.

        Raises:
            TranspilerError: if the coupling map or the layout are not
            compatible with the DAG.
        """
        if isinstance(dag, DAGCircuit):
            self._check_qiskit(dag)
            circuit = from_qiskit(dag)
        else:
            circuit = dag

        if self.method == "sabre":
            mapped_circuit, _ = sabre_map(self.device, circuit)
        elif self.method == "lazy":
            mapped_circuit, _ = jit_map(self.device, circuit)
        elif self.method == "bridge":
            mapped_circuit, _ = bridge_map(self.device, circuit)

        if isinstance(dag, DAGCircuit):
            return to_qiskit(mapped_circuit)
        return mapped_circuit

    def _check_qiskit(self, dag: DAGCircuit):
        if len(dag.qregs) != 1 or dag.qregs.get("q", None) is None:
            raise TranspilerError("Mapping runs on physical circuits only")



# ------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
# ------------------------------------------------------------------------------
from tweedledum.ir import Circuit


def test_wires():
    """Adding Cbits and Qubits to a circuit."""
    circuit = Circuit()
    q0 = circuit.create_qubit()
    c0 = circuit.create_cbit()
    c1 = circuit.create_cbit()
    q1 = circuit.create_qubit()

    assert circuit.num_cbits() == 2
    assert circuit.num_qubits() == 2
    assert circuit.num_wires() == 4
    assert q0.uid() == c0.uid()
    assert q1.uid() == c1.uid()
    assert q0.uid() != c1.uid()



# -------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
# -------------------------------------------------------------------------------
import _ast
import ast

from .bitvec import BitVec
from tweedledum.classical import LogicNetwork


class ParseError(Exception):
    pass


class FunctionParser(ast.NodeVisitor):
    bit_ops = {
        _ast.BitAnd: "create_and",
        _ast.BitOr: "create_or",
        _ast.BitXor: "create_xor",
    }

    bool_ops = {
        _ast.And: "create_and",
        _ast.Or: "create_or",
    }

    # This feels quite hack-y
    types = {"BitVec": type(BitVec(1))}

    def __init__(self, source):
        self._symbol_table = dict()
        self._parameters_signature = list()
        self._return_signature = list()
        self._logic_network = LogicNetwork()

        node = ast.parse(source)
        self.visit(node)
        super().__init__()

    def visit_args(self, node):
        cache = list()
        for arg in node.args:
            cache.append(arg.arg)
            if arg.annotation is None:
                continue
            if not isinstance(arg.annotation, _ast.Call):
                raise ParseError("BitVec type with size is needed")
            for var in cache:
                pis = list()
                arg_type, arg_size = self._visit_annotation_Call(arg.annotation)
                for i in range(arg_size):
                    pis.append(self._logic_network.create_pi("{}_{}".format(var, i)))
                self._symbol_table[var] = ((arg_type, arg_size), pis)
                self._parameters_signature.append((arg_type, arg_size))
            cache.clear()
        if len(cache) != 0:
            raise ParseError("Argument type is needed for %s" % cache)

    def visit_Assign(self, node):
        """When assign, the scope needs to be updated with the right type"""
        value_type, value_signals = self.visit(node.value)
        for target in node.targets:
            self._symbol_table[target.id] = (value_type, value_signals)
        return (value_type, value_signals)

    def visit_BinOp(self, node):
        """Handles ``&``, ``^``, and ``|``."""
        op = FunctionParser.bit_ops.get(type(node.op))
        if not op:
            raise ParseError("Unknown binop.op %s" % op)
        left_type, left_signals = self.visit(node.left)
        right_type, right_signals = self.visit(node.right)
        if len(left_signals) != len(right_signals):
            raise ParseError("Different length binop.op %s" % op)
        result = list()
        for l, r in zip(left_signals, right_signals):
            result.append(getattr(self._logic_network, op)(l, r))
        return right_type, result

    def visit_BoolOp(self, node):
        _, result_signal = self.visit(node.values[0])
        op = FunctionParser.bool_ops.get(type(node.op))
        for value in node.values[1:]:
            _, value_signal = self.visit(value)
            result_signal[0] = getattr(self._logic_network, op)(
                result_signal[0], value_signal[0]
            )
        return (FunctionParser.types["BitVec"], 1), result_signal

    def visit_Call(self, node):
        type_ = FunctionParser.types[node.func.id]
        if len(node.args) == 1:
            value = ast.literal_eval(node.args[0])
            if isinstance(value, str):
                return (type_, len(value)), self._const_BitVec(len(value), value)
            elif isinstance(value, int):
                return (type_, value), self._const_BitVec(value)
        elif len(node.args) == 2:
            length = ast.literal_eval(node.args[0])
            if not isinstance(length, int):
                raise ParseError("BitVec requires length to be an integer")
            value = ast.literal_eval(node.args[1])
            if isinstance(value, int):
                value = "{:0{}b}".format(value, length)
            return (type_, length), self._const_BitVec(length, value)
        else:
            raise ParseError("Invalid number of arguments")

    def visit_Compare(self, node):
        left_type, left_signals = self.visit(node.left)
        partial_results = list()
        for op, right in zip(node.ops, node.comparators):
            right_type, right_signals = self.visit(right)
            if len(left_signals) != len(right_signals):
                raise ParseError("Different length")
            if isinstance(op, ast.Eq):
                new_signals = [
                    self._logic_network.create_xnor(i, j)
                    for i, j in zip(left_signals, right_signals)
                ]
                partial_results.append(self._logic_network.create_nary_and(new_signals))
            elif isinstance(op, ast.NotEq):
                new_signals = [
                    self._logic_network.create_xor(i, j)
                    for i, j in zip(left_signals, right_signals)
                ]
                partial_results.append(self._logic_network.create_nary_or(new_signals))
            else:
                raise ParseError("Unsupported comparator")

        result = self._logic_network.create_nary_and(partial_results)
        return (FunctionParser.types["BitVec"], 1), [result]

    def visit_FunctionDef(self, node):
        if node.returns is None:
            raise ParseError("Return type is needed")
        if isinstance(node.returns, _ast.Call):
            return_type = node.returns.func.id
            size = ast.literal_eval(node.returns.args[0])
            if not isinstance(size, int):
                ParseError("Size must be an integer")
            self._return_signature = [(FunctionParser.types[return_type], size)]
            self._symbol_table["__dee_ret_0"] = (return_type, None)
        elif isinstance(node.returns, ast.Tuple):
            for i, elt in enumerate(node.returns.elts):
                return_type = elt.func.id
                size = ast.literal_eval(elt.args[0])
                if not isinstance(size, int):
                    ParseError("Size must be an integer")
                self._return_signature.append((FunctionParser.types[return_type], size))
                self._symbol_table[f"__dee_ret_{i}"] = (return_type, None)
        else:
            raise ParseError("Return type is needed")

        self.visit_args(node.args)
        return super().generic_visit(node)

    def visit_Name(self, node):
        if node.id not in self._symbol_table:
            raise ParseError("out of scope: %s" % node.id)
        return self._symbol_table[node.id][0], self._symbol_table[node.id][1]

    def visit_Return(self, node):
        """The return type should match the return type hint."""
        if isinstance(node.value, ast.Tuple):
            if len(self._return_signature) != len(node.value.elts):
                raise ParseError(
                    f"The function was expected to return "
                    f"{len(self._return_signature)} values, "
                    f"but it returned {len(node.value.elts)}."
                )
            for i, elt in enumerate(node.value.elts):
                elt_type, elt_signals = self.visit(elt)
                if elt_type != self._return_signature[i]:
                    raise ParseError(
                        f"The {i}-th return value was expected to "
                        f"be of type {self._return_signature[i]}, "
                        f"but got {elt_type}."
                    )

                for s in elt_signals:
                    self._logic_network.create_po(s)
        else:
            if len(self._return_signature) > 1:
                raise ParseError(
                    f"The function was expected to return "
                    f"{len(self._return_signature)} values, "
                    f"but it returned just 1."
                )
            return_type, signals = self.visit(node.value)
            if return_type != self._return_signature[0]:
                raise ParseError(
                    f"The return value was expected to "
                    f"be of type {self._return_signature[0]}, "
                    f"but got {return_type}."
                )
            for s in signals:
                self._logic_network.create_po(s)

    def visit_Slice(self, node):
        lower = ast.literal_eval(node.lower)
        upper = ast.literal_eval(node.lower)
        return slice(lower, upper)

    # Python < 3.9
    def visit_Index(self, node):
        if not isinstance(node.value, ast.Num):
            raise ParseError("Subscript index must be a number")
        return ast.literal_eval(node.value)

    def visit_Constant(self, node):
        return ast.literal_eval(node)

    def visit_Subscript(self, node):
        v_type, v_signals = self.visit(node.value)
        slice_ = self.visit(node.slice)
        if isinstance(slice_, int):
            return v_type, [v_signals[slice_]]
        if isinstance(slice_, slice):
            return v_type, v_signals[slice_]
        raise ParseError("Subscript must be an integer or a slice")

    def visit_UnaryOp(self, node):
        result_type, result_signal = self.visit(node.operand)
        if isinstance(node.op, _ast.Not):
            if len(result_signal) != 1:
                raise ParseError("Boolean NOT doesn't work on multibit values")
            return result_type, [self._logic_network.create_not(result_signal[0])]
        elif isinstance(node.op, ast.Invert):
            result = list()
            for s in result_signal:
                result.append(self._logic_network.create_not(s))
            return result_type, result
        raise ParseError("Unsupported unary operator")

    def _visit_annotation_Call(self, node):
        type_ = FunctionParser.types[node.func.id]
        if len(node.args) == 1:
            size = ast.literal_eval(node.args[0])
            if not isinstance(size, int):
                ParseError("Size must be an integer")
            return type_, size
        raise ParseError("Invalid number of arguments")

    def _const_BitVec(self, length, value=None):
        if value == None:
            return [self._logic_network.get_constant(0) for i in range(length)]
        if not isinstance(value, str):
            raise ParseError(
                "When creating constant BitVec, value must be a string or None"
            )
        if len(value) > length:
            ParseError(
                f"BitVec value requires a bit vector of "
                f"length {len(value)}, but declared BitVec has "
                f"length {length}"
            )
        return [self._logic_network.get_constant(i == "1") for i in value[::-1]]



# -------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
# -------------------------------------------------------------------------------
from typing import Union, List
from os.path import isfile
import inspect
import math
import types

from .bitvec import BitVec
from .function_parser import FunctionParser
from .._tweedledum.classical import TruthTable, create_from_binary_string
from .._tweedledum import classical


class BoolFunction(object):
    """Class to represent a Boolean function

    Formally, a Boolean function is a mapping :math:`f : {0, 1}^n \to {0, 1}^m`,
    where :math:`n`(:math:`m`) is the number of inputs(outputs).  There are
    many ways to represent/specify a Boolean function.  Here, we use two:

    Truth tables:  They are are an explicit and _not compact_ function
        representation.  Basically a truth table is an exhaustive mapping from
        input binary bit-strings of length :math:`n` to corresponding output
        bit-strings of length :math:`m`. Hence they do not scale well. They are,
        however, tremendously useful to represent and manipulate small functions.

    Logic network (Xor-And graph, XAG): A logic network is modeled by a directed
        acyclic graph where nodes represent primary inputs and outputs, as well
        as local functions.  The nodes representing local function as called
        gates.  In our case, we limit our local functions to be either a 2-input
        AND or a 2-input XOR---a structure known as XAG.  Therefore, a XAG is a
        2-regular non-homogeneous logic network.

    Under the hood both representations are implemented in C++.
    """

    def __init__(self, f):
        if not isinstance(f, types.FunctionType):
            raise TypeError("Constructor requires a function")
        parsed_function = FunctionParser(inspect.getsource(f).strip())
        self._parameters_signature = parsed_function._parameters_signature
        self._return_signature = parsed_function._return_signature
        self._logic_network = parsed_function._logic_network
        self._truth_table = None
        self._num_input_bits = self._logic_network.num_pis()
        self._num_output_bits = self._logic_network.num_pos()

    def _format_simulation_result(self, sim_result):
        i = 0
        result = list()
        for type_, size in self._return_signature:
            tmp = sim_result[i : i + size]
            result.append(type_(size, tmp[::-1]))
            i += size
        if len(result) == 1:
            return result[0]
        return tuple(result)

    def num_inputs(self):
        return len(self._parameters_signature)

    def num_outputs(self):
        return len(self._return_signature)

    def num_input_bits(self):
        return self._num_input_bits

    def num_output_bits(self):
        return self._num_output_bits

    def simulate(self, *argv):
        if len(argv) != self.num_inputs():
            raise RuntimeError(
                f"The function requires {self.num_inputs()}. "
                f"It's signature is: {self._parameters_signature}"
            )
        input_str = str()
        for i, arg in enumerate(argv):
            arg_type = (type(arg), len(arg))
            if arg_type != self._parameters_signature[i]:
                raise TypeError(
                    f"Wrong argument type. Argument {i} "
                    f"expected: {self._parameters_signature[i]}, "
                    f"got: {arg_type}"
                )
            arg_str = str(arg)
            input_str += arg_str[::-1]

        # If the truth table was already computed, we just need to look for the
        # result of this particular input
        if self._truth_table != None:
            position = int(input_str[::-1], base=2)
            sim_result = "".join([str(int(tt[position])) for tt in self._truth_table])
        else:
            input_vector = [bool(int(i)) for i in input_str]
            sim_result = classical.simulate(self._logic_network, input_vector)
            sim_result = "".join([str(int(i)) for i in sim_result])

        return self._format_simulation_result(sim_result)

    def simulate_all(self):
        if self._truth_table == None:
            self._truth_table = classical.simulate(self._logic_network)

        result = list()
        for position in range(2 ** self._logic_network.num_pis()):
            sim_result = "".join([str(int(tt[position])) for tt in self._truth_table])
            result.append(self._format_simulation_result(sim_result))

        return result

    def logic_network(self):
        return self._logic_network

    def truth_table(self, output_bit: int):
        if not isinstance(output_bit, int):
            raise TypeError("Parameter output must be an integer")
        if self._truth_table == None:
            self.simulate_all()
        return self._truth_table[output_bit]

    @classmethod
    def from_expression(cls, expression: str, var_order: list = None):
        """
        Create a BooleanFunction from an arbitrary logical expression.

        A logical expression is composed of logical operators & (AND), | (OR),
        ~ (NOT), and ^ (XOR), as well as symbols for variables. For example,
        'a & b' and '~(~(x0 | x1) ^ (~x0 & ~x1))' are both valid string
        representation of logical expressions.

        Logic expressions are parsed and stored, initially, as a LogicNetwork,
        which is a XOR-AND graph (XAG).

        Args:
            expression: The string of the desired logical expression, such as
                        'a & b' or '~(~(x0 | x1) ^ (~x0 & ~x1))'

        Returns:
            A BooleanFunction that uses a XAG to represent the logical expression
        """
        from .expression_parser import ExpressionParser

        function = cls.__new__(cls)
        parsed_expression = ExpressionParser(expression, var_order)
        function._parameters_signature = parsed_expression._parameters_signature
        function._return_signature = parsed_expression._return_signature
        function._logic_network = parsed_expression._logic_network
        function._truth_table = None
        function._num_input_bits = len(function._parameters_signature)
        function._num_output_bits = 1
        return function

    @classmethod
    def from_function(cls, function):
        """Create a BooleanFunction from a python function.

        Returns:
            A BooleanFunction for the python function
        """
        return BoolFunction(function)

    @classmethod
    def from_truth_table(cls, tts: Union[str, List[str]]):
        """Create a BooleanFunction from an truth table string.

        Truth tables are a common way of specifying boolean functions. They are
        are an explicit and _not compact_ function representation---basically a
        truth table is an exhaustive mapping from input binary bit-strings of
        length :math:`n` to corresponding output bit-strings of length :math:`m`.
        However, they are tremendously useful to represent and manipulate small
        functions.  For example, the following is a simple truth table that
        corresponds to the `AND` of two variables:

        =====  =====  =============
           Inputs        Output
        ------------  -------------
          a      b      f(a, b)
        =====  =====  =============
          0      0         0
          0      1         0
          1      0         0
          1      1         1
        =====  =====  =============

        In this case :math:`n=2`, and :math:`m=1`.  Note that, for compactness,
        the input bit-strings are omitted because they can be easily derived
        for any given :math:`n`.  Hence to completely specify a truth table,
        we only need a Length-2 :sup:`n` bit-string for each of the :math:`m`
        outputs.  In the above example, a single bit-string `'1000'` is enough.
        The most-significant bit corresponds to the :math:`f(1, 1)` and the
        least-significant bit corresponds to :math:`f(0, 0)`.

        Args:
            tt: A single binary string or a list of binary strings representing
                the a single- or multi-output truth table.

        Returns:
            A BooleanFunction that uses truth tables to represent the function
        """
        if isinstance(tts, str):
            tts = [tts]

        # Check that the input binary stings have the same length and that this
        # length is a power of 2
        if not ((len(tts[0]) & (len(tts[0]) - 1) == 0) and len(tts[0]) != 0):
            raise RuntimeError("Length of all binary string must be a power of 2")
        for tt in tts[1:]:
            if len(tt) != len(tts[0]):
                raise RuntimeError("Length of all binary string must equal")

        n_inputs = int(math.log(len(tts[0]), 2))
        n_outputs = len(tts)

        function = cls.__new__(cls)
        function._logic_network = None
        function._truth_table = [TruthTable(n_inputs) for _ in range(n_outputs)]
        for i, tt in enumerate(tts):
            create_from_binary_string(function._truth_table[i], tt)
        function._parameters_signature = [(type(BitVec(1)), 1)] * n_inputs
        function._return_signature = [(type(BitVec(1)), 1)] * n_outputs
        function._num_input_bits = n_inputs
        function._num_output_bits = n_outputs
        return function

    @classmethod
    def from_aiger_file(cls, path: str):
        """Create a BooleanFunction from an AIGER file.

        AIGER is a binary format to store Boolean functions that are represented
        as And-Inverter Graphs (AIGs).  An AIG is a 2-regular homogeneous logic
        network, in which each gate function is the 2-input AND. Inverters are
        modelled using the complementation flags.

        The logic network used in `BoolFunction` is _not_ an AIG.  We use a XAG
        instead.  A XAG is a 2-regular non-homogeneous logic network, in which
        each gate function is either a 2-input AND or a 2-input XOR.  Inverters
        are modelled using the complementation flags, but they are not strictly
        necessary.  Thus this function reads an AIG from a AIGER file and
        transforms it into a XAG.

        Args:
            path: Path to the AIGER file (*.aig)

        Returns:
            A BooleanFunction that uses a XAG to represent the function
        """
        if not isfile(path):
            raise FileNotFoundError(f"The file {path} does not exists.")
        function = cls.__new__(cls)
        function._logic_network = classical.read_aiger(path)
        function._truth_table = None
        n_inputs = function._logic_network.num_pis()
        n_outputs = function._logic_network.num_pos()
        function._parameters_signature = [(type(BitVec(1)), 1)] * n_inputs
        function._return_signature = [(type(BitVec(1)), 1)] * n_outputs
        function._num_input_bits = n_inputs
        function._num_output_bits = n_outputs
        return function

    @classmethod
    def from_dimacs_file(cls, path: str):
        """Create a BooleanFunction from a DIMACS CNF file.

        The `DIMACS CNF format
        <http://www.satcompetition.org/2009/format-benchmarks2009.html>`__, is
        the standard format for specifying SATisfiability (SAT) problem
        instances in `Conjunctive Normal Form (CNF)
        <https://en.wikipedia.org/wiki/Conjunctive_normal_form>`__, which is a
        two-level representation of a Boolean function, in which the inner
        operator is OR and the outter operator is AND.  Thus a CNF is a
        conjunction of one or more clauses, where a clause is a disjunction of
        one or more literals.

        Args:
            path: Path to the DIMACS file (*.cnf)

        Returns:
            A BooleanFunction that uses a XAG to represent the function
        """
        if not isfile(path):
            raise FileNotFoundError(f"The file {path} does not exists.")
        function = cls.__new__(cls)
        function._logic_network = classical.read_dimacs(path)
        function._truth_table = None
        n_inputs = function._logic_network.num_pis()
        n_outputs = function._logic_network.num_pos()
        function._parameters_signature = [(type(BitVec(1)), 1)] * n_inputs
        function._return_signature = [(type(BitVec(1)), 1)] * n_outputs
        function._num_input_bits = n_inputs
        function._num_output_bits = n_outputs
        return function

    @classmethod
    def from_verilog_file(cls, path: str):
        """Create a BooleanFunction from Verilog file.

        /!\ This function can only parse a _very_ small and simple subset of
        /!\ the Verilog language.

        Verilog is a hardware description language (HDL) used to model
        electronic systems.  Here we are interested on its capability to model
        purely combinational circuits.

        Args:
            path: Path to the Verilog file (*.v)

        Returns:
            A BooleanFunction that uses a XAG to represent the function
        """
        if not isfile(path):
            raise FileNotFoundError(f"The file {path} does not exists.")
        function = cls.__new__(cls)
        function._logic_network = classical.read_verilog(path)
        function._truth_table = None
        n_inputs = function._logic_network.num_pis()
        n_outputs = function._logic_network.num_pos()
        function._parameters_signature = [(type(BitVec(1)), 1)] * n_inputs
        function._return_signature = [(type(BitVec(1)), 1)] * n_outputs
        function._num_input_bits = n_inputs
        function._num_output_bits = n_outputs
        return function



# -------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
# -------------------------------------------------------------------------------
from typing import Union

from qiskit.dagcircuit import DAGCircuit
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.coupling import CouplingMap
from qiskit.transpiler.exceptions import TranspilerError

from tweedledum.qiskit import from_qiskit, to_qiskit
from tweedledum.ir import Circuit
from tweedledum.passes import steiner_resynth
from tweedledum.target import Device


class SteinerResynth(TransformationPass):
    """TODO"""

    def __init__(self, cmap_or_device: Union[list, CouplingMap, Device]):
        """SteinerResynth initializer.

        Args:
            cmap_or_device (Union[list, CouplingMap, Device]): Directed graph
                represented as a list of edges, or a qiskit coupling map, or a
                tweedledum Device.
        """
        super().__init__()
        if isinstance(cmap_or_device, list):
            self.device = Device.from_edge_list(cmap_or_device)
        elif isinstance(cmap_or_device, CouplingMap):
            self.device = Device.from_edge_list(cmap_or_device.get_edges())
        elif isinstance(cmap_or_device, Device):
            self.device = cmap_or_device
        else:
            raise TranspilerError("Must pass a qiskit CouplingMap or tweedledum Device")

    def run(self, dag: Union[DAGCircuit, Circuit]) -> Union[DAGCircuit, Circuit]:
        """Run the LinearResynth pass on a circuit.

        Args:
            dag (Union[DAGCircuit, Circuit]): A circuit to map.

        Returns:
            Union[DAGCircuit, Circuit]: An optimized circuit.
        """
        if isinstance(dag, DAGCircuit):
            circuit = from_qiskit(dag)
        else:
            circuit = dag

        opt_circuit = steiner_resynth(self.device, circuit)

        if isinstance(dag, DAGCircuit):
            return to_qiskit(opt_circuit)
        return opt_circuit



# -------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
# -------------------------------------------------------------------------------
import os
import sys
import skbuild
import setuptools

build_dev_version = os.environ.get("TWEEDLEDUM_DEV_VERSION", False)


def get_name():
    if build_dev_version:
        return "tweedledum-dev"
    else:
        return "tweedledum"


def get_version():
    if build_dev_version:
        return "1.2.0.dev" + str(build_dev_version)
    else:
        return "1.1.0"


skbuild.setup(
    name=get_name(),
    version=get_version(),
    package_dir={"": "python"},
    packages=setuptools.find_packages(
        where="python", include=["tweedledum", "tweedledum.*"]
    ),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.6",
    cmake_install_dir="python",
)

# /!\ This is a 'gambiarra':
# For whatever reason, whenever I try to install `tweedledum` in development
# mode, i.e., `pip install -e .`, the easy_path and egg-link are wrong! Hence,
# I basically fix them by hand because I couldn't figure out a better way of
# doing it!
if len(sys.argv) > 1 and sys.argv[1] == "develop":
    from pathlib import Path
    import fileinput
    import shutil
    import site

    easy_install = Path(site.getsitepackages()[0]) / "easy-install.pth"
    egg_link = Path(site.getsitepackages()[0]) / "tweedledum.egg-link"
    for line in fileinput.input([easy_install, egg_link], inplace=True):
        print(line.replace("tweedledum", "tweedledum/python"), end="")
    shutil.rmtree("tweedledum.egg-info", ignore_errors=True)



# -------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
# -------------------------------------------------------------------------------
from tweedledum.ir import Circuit, Qubit, Cbit, rotation_angle
from tweedledum import operators as Op

from qiskit.circuit import (
    QuantumCircuit,
    QuantumRegister,
    ClassicalRegister,
    Gate,
    ControlledGate,
    Measure,
)
from qiskit.dagcircuit.dagcircuit import DAGCircuit
from qiskit.circuit.library.standard_gates import (
    HGate,
    IGate,
    PhaseGate,
    RXGate,
    RYGate,
    RZGate,
    RXXGate,
    RYYGate,
    RZZGate,
    SGate,
    SdgGate,
    SXGate,
    SXdgGate,
    SwapGate,
    TGate,
    TdgGate,
    XGate,
    YGate,
    ZGate,
)

_TO_TWEEDLEDUM_OP = {
    "h": Op.H,
    "p": Op.P,
    "rx": Op.Rx,
    "ry": Op.Ry,
    "rz": Op.Rz,
    "rxx": Op.Rxx,
    "ryy": Op.Ryy,
    "rzz": Op.Rzz,
    "s": Op.S,
    "sdg": Op.Sdg,
    "sx": Op.Sx,
    "sxdg": Op.Sxdg,
    "swap": Op.Swap,
    "t": Op.T,
    "tdg": Op.Tdg,
    "x": Op.X,
    "y": Op.Y,
    "z": Op.Z,
    "measure": Op.Measure,
}

_TO_QISKIT_GATE = {
    "std.h": HGate,
    "std.p": PhaseGate,
    "std.m": Measure,
    "std.rx": RXGate,
    "std.ry": RYGate,
    "std.rz": RZGate,
    "std.s": SGate,
    "std.sdg": SdgGate,
    "std.sx": SXGate,
    "std.sxdg": SXdgGate,
    "std.swap": SwapGate,
    "std.t": TGate,
    "std.tdg": TdgGate,
    "std.x": XGate,
    "std.y": YGate,
    "std.z": ZGate,
    "ising.rxx": RXXGate,
    "ising.ryy": RYYGate,
    "ising.rzz": RZZGate,
}


def _convert_qiskit_operator(gate):
    op = _TO_TWEEDLEDUM_OP.get(gate.name) or _TO_TWEEDLEDUM_OP.get(gate.base_gate.name)
    ctrl_state = ""
    if isinstance(gate, ControlledGate):
        ctrl_state = "{:0{}b}".format(gate.ctrl_state, gate.num_ctrl_qubits)[::-1]
    if op == None:
        return gate, ctrl_state
    if isinstance(gate.params, list) and len(gate.params) > 0:
        if len(gate.params) == 1:
            return op(gate.params[0]), ctrl_state
        else:
            return gate, ctrl_state
    return op(), ctrl_state


def _from_qc(qiskit_qc):
    circuit = Circuit()
    qubits = [circuit.create_qubit() for i in range(len(qiskit_qc.qubits))]
    qubits_map = dict(zip(qiskit_qc.qubits, qubits))
    cbits = [circuit.create_cbit() for i in range(len(qiskit_qc.clbits))]
    cbits_map = dict(zip(qiskit_qc.clbits, cbits))

    for instruction, qargs, cargs in qiskit_qc.data:
        op, ctrl_state = _convert_qiskit_operator(instruction)
        qs = [qubits_map.get(qubit) for qubit in qargs]
        cs = [cbits_map.get(cbit) for cbit in cargs]
        for i, polarity in enumerate(ctrl_state):
            if polarity == "0":
                qs[i] = ~qs[i]
        circuit.apply_operator(op, qs, cs)
    return circuit


def _from_dag(qiskit_dag):
    circuit = Circuit()
    qubits = [circuit.create_qubit() for i in range(len(qiskit_dag.qubits))]
    qubits_map = dict(zip(qiskit_dag.qubits, qubits))
    cbits = [circuit.create_cbit() for i in range(len(qiskit_dag.clbits))]
    cbits_map = dict(zip(qiskit_dag.clbits, cbits))

    for node in qiskit_dag.op_nodes():
        op, ctrl_state = _convert_qiskit_operator(node.op)
        qs = [qubits_map.get(qubit) for qubit in node.qargs]
        cs = [cbits_map.get(cbit) for cbit in node.cargs]
        for i, polarity in enumerate(ctrl_state):
            if polarity == "0":
                qs[i] = ~qs[i]
        circuit.apply_operator(op, qs, cs)
    return circuit


def _convert_tweedledum_op(op):
    base_gate = _TO_QISKIT_GATE.get(op.kind())
    if base_gate == None:
        if op.kind() == "py_op":
            return op.py_op()
        else:
            raise RuntimeError(f"Unrecognized operator: {op.kind()}")

    if op.kind() in [
        "std.p",
        "std.rx",
        "std.ry",
        "std.rz",
        "ising.rxx",
        "ising.ryy",
        "ising.rzz",
    ]:
        angle = rotation_angle(op)
        gate = base_gate(angle)
    else:
        gate = base_gate()

    # TODO: need to deal with cbits too!
    if op.num_controls() > 0:
        qubits = op.qubits()
        ctrl_state = ""
        for qubit in qubits[: op.num_controls()]:
            ctrl_state += "{}".format(int(qubit.polarity() == Qubit.Polarity.positive))
        return gate.control(len(ctrl_state), ctrl_state=ctrl_state[::-1])
    return gate


def _to_qc(circuit):
    qiskit_qc = QuantumCircuit()
    # Qiskit is weird, it doesn't allow passing 0 as size for register.
    if circuit.num_cbits():
        qiskit_qc.add_register(circuit.num_qubits(), circuit.num_cbits())
    else:
        qiskit_qc.add_register(circuit.num_qubits())

    for instruction in circuit:
        gate = _convert_tweedledum_op(instruction)
        qubits = [qubit.uid() for qubit in instruction.qubits()]
        cbits = [cbit.uid() for cbit in instruction.cbits()]
        qiskit_qc.append(gate, qubits, cbits)
    return qiskit_qc


def _to_dag(circuit):
    qiskit_dag = DAGCircuit()
    qreg = QuantumRegister(circuit.num_qubits())
    qiskit_dag.add_qreg(qreg)
    if circuit.num_cbits():
        creg = ClassicalRegister(circuit.num_cbits())
        qiskit_dag.add_creg(creg)
    for instruction in circuit:
        gate = _convert_tweedledum_op(instruction)
        qubits = [qreg[qubit.uid()] for qubit in instruction.qubits()]
        cbits = [creg[cbit.uid()] for cbit in instruction.cbits()]
        qiskit_dag.apply_operation_back(gate, qubits, cbits)
    return qiskit_dag


def from_qiskit(circuit):
    if isinstance(circuit, QuantumCircuit):
        return _from_qc(circuit)
    elif isinstance(circuit, DAGCircuit):
        return _from_dag(circuit)
    raise TypeError("Circuit must be either a QuantumCircuit or DAGCircuit")


def to_qiskit(circuit, circuit_type="dag"):
    if circuit_type == "dag":
        return _to_dag(circuit)
    elif circuit_type == "gatelist":
        return _to_qc(circuit)
    raise TypeError(f"Unrecognized circuit type {circuit_type}")



#-------------------------------------------------------------------------------
# Part of Tweedledum Project.  This file is distributed under the MIT License.
# See accompanying file /LICENSE for details.
#-------------------------------------------------------------------------------
class BitVec(object):
    """Class to compactly represent a vector of bits

    The BitVec type is an unconstrained vector. During its declaration the
    length must be specified and cannot be modified later.

    The class implements an indexing interface that provides access to the bits
    of the integer representation. Hence, assignments to a BitVec can be done 
    using single element assignments or slices. Note that lower indices 
    correspond to less significant bits. (Little endian: bn .. b3 b2 b1 b0).
    Example:
        >>> x = BitVec(4, '0101')
        >>> x[0]
        1
        >>> x[1]
        0
        >>> x[2]
        1
        >>> x[3]
        0

    Unlike standard Python, slicing ranges are downward. Thus, the highest index
    bit corresponds to the leftmost item. As in standard Python, the slicing
    range is half-open: this highest index bit is not included.
    Example:
        >>> x = BitVec(4, '0101')
        >>> x[3:0]
        '101'
        >>> x[3:1]
        '10'
        >>> x[4:0]
        '0101'
    
    Both indices can be omitted from the slice. If the leftmost index is
    omitted, the meaning is to access “all” higher order bits.  If the rightmost
    index is omitted, it is 0 by default. 
    Example:
        >>> x = BitVec(4, '0101')
        >>> x[:1]
        '010'
        >>> x[2:]
        '01'

    Attributes
    ----------
    _length : int
        length of the vector 
    _value : int
        an integer used to compactly store the bits
    """

    def __init__(self, length, value = 0):
        """
        Parameters
        ----------
        length : int
            length of the vector 
        value : int or str (bit string), optional
            the initial value (default = 0)
        """
        if isinstance(length, str) and value == 0:
            value = length
            length = len(value)
        if not isinstance(length, int):
            raise TypeError("[BitVec] The length must be an integer value")
        self._length = length
        if isinstance(value, int):
            # this -2 account for the leading '0b': 0b....
            # FIXME: feels a bit hack-y, because I suck at python :(
            required_length = len(bin(value)) - 2 
            if required_length > length:
                raise TypeError(f"[BitVec] Value requires a bit vector of "
                                f"length {required_length}, but BitVec has "
                                f"length {length}")
            self._value = value
        elif isinstance(value, str):
            if (len(value) > length):
                raise TypeError(f"[BitVec] String requires a bit vector of "
                                f"length {len(value)}, but BitVec has length " 
                                f"{length}")
            self._value = int(value, base = 2)
        else:
            raise TypeError("[BitVec] The value must be either an interger or "
                            "a bit string")

    def __len__(self):
        return self._length

    def __bool__(self):
        if self._length > 1:
            raise TypeError("[BitVec] A BitVec of length bigger than one "
                            "cannot be converted to a Boolean value")
        return bool(self._value)

    def __invert__ (self):
        return BitVec(self._length, ~self._value & (1 << self._length) - 1)

    def __and__(self, other):
        if isinstance(other, BitVec):
            if (self._length != other._length):
                raise TypeError("[BitVec] __and__ operation: length mismatch")
            return BitVec(self._length, self._value & other._value)
        raise TypeError("[BitVec] __and__ operation: type mismatch")

    def __or__(self, other):
        if isinstance(other, BitVec):
            if (self._length != other._length):
                raise TypeError("[BitVec] __or__ operation: length mismatch")
            return BitVec(self._length, self._value | other._value)
        raise TypeError("[BitVec] __or__ operation: type mismatch")

    def __xor__(self, other):
        if isinstance(other, BitVec):
            if (self._length != other._length):
                raise TypeError("[BitVec] __xor__ operation: length mismatch")
            return BitVec(self._length, self._value ^ other._value)
        raise TypeError("[BitVec] __xor__ operation: type mismatch")

    def __eq__(self, other):
        if isinstance(other, BitVec):
            return BitVec(1, self._length == other._length and self._value == other._value)
        raise TypeError("__eq__ operation: type mismatch")

    def __ne__(self, other):
        if isinstance(other, BitVec):
            return BitVec(1, self._length != other._length or self._value != other._value)
        raise TypeError("__ne__ operation: type mismatch")

    def __repr__(self):
        return f"BitVec({self._length}, '{str(self)}')"

    def __str__(self):
        return "{:0{}b}".format(self._value, self._length)

    def __getitem__(self, index):
        if isinstance(index, slice):
            i, j = index.start, index.stop
            if j is None:
                j = 0
            if j < 0:
                raise ValueError("[BitVec] Slice [i:j] requires j >= 0\n")
            if i is None:
                return BitVec(j, self._value >> j)
            if i <= j:
                raise ValueError("[BitVec] Slice [i:j] requires i > j\n")
            return BitVec(i - j, (self._value & (1 << i) - 1) >> j)
        else:
            return BitVec(1, int((self._value >> index) & 0x1))

    def __setitem__(self, index, value):
        if isinstance(index, slice):
            if not isinstance(value, BitVec):
                raise TypeError("[BitVec] Assignment operation: type mismatch")
            i, j = index.start, index.stop
            if j is None or j < 0:
                j = 0
            if i is None:
                i = self._length
            if i <= j:
                raise ValueError("[BitVec] Slice [i:j] requires i > j\n")
            if (i - j) != value._length:
                raise ValueError("[BitVec] Assignment requires BitVec of with "
                                 "the same length\n")
            limit = (1 << (i - j))
            mask = (limit - 1) << j
            self._value &= ~mask
            self._value |= (value._value << j)
        else:
            if isinstance(value, BitVec):
                if value._length != 1:
                    raise ValueError("[BitVec] Single element assignment "
                                     "requires an interger value (0 or 1) or a "
                                     "BitVec of length one")
                v = value._value
            elif isinstance(value, int):
                v = value
            else:
                raise ValueError("[BitVec] Single element assignment requires "
                                 "an interger value (0 or 1) or a "
                                 "BitVec of length one")
            if v == 1:
                self._value |= (1 << index)
            elif v == 0:
                self._value &= ~(1 << index)
            else:
                raise ValueError("[BitVec] Single element assignment requires "
                                 "an interger value (0 or 1) or a "
                                 "BitVec of length one")
